#########################################################
# Lilith
#
# Routine example to perform a coupling fit
# Use the Minuit library iminuit to perform minimization
# Model: (C_gamma, C_g)
#
#########################################################

import lilith, tools.genXML
from iminuit import Minuit

# Use the library matplotlib to plot the results
# To comment if not installed
from pylab import *
import matplotlib.pyplot as plt


######################################################################

# Choose which experimental data to fit
myexpinput = "data/ATLASdiboson.list"
# Lilith precision mode
myprecision = "BEST-QCD"
# Higgs mass to test
myhmass = 125.5
# Output file name (plots + dat file)
outputfile = 'results/1D_ATLAS_CGaCg'
verbose=False

######################################################################



######################################################################

def getL(CGa, Cg):
  """ Returns -2Log(L) for given (C_gamma, C_gluon), based on Lilith classes and functions """
  myXML_user_input = tools.genXML.redC(myhmass, CGaGa=CGa, Cgg=Cg, precision=myprecision)
  l = Lilithcalc.GetLikelihood(myXML_user_input)[0]
  return l

######################################################################

# Initialize a Lilith object
Lilithcalc = lilith.Lilith(verbose)
# Read experimental data
Lilithcalc.SetExpInput(myexpinput)

# Initialize the fit; parameter starting values and limits
m = Minuit(getL, CGa=0.9, limit_CGa=(0,2), Cg=0.9, limit_Cg=(0,2))
print m.print_param()
# Fit the model
m.migrad()
# Display parameter values at the best-fit point
print "C_gamma:", m.values["CGa"], " C_gluon:", m.values["Cg"], " -2Log(L) min:", m.fval

# Get the 1D profile for the parameters
nbins = 250
xg,yg,rg = m.mnprofile('Cg', bins=nbins, bound=(0.0, 2.0), subtract_min=True)
xGa,yGa,rGa = m.mnprofile('CGa', bins=nbins, bound=(0.0, 2.0), subtract_min=True)


# Save results in output file
f=open(outputfile+".dat",'w')
for i in range(1, nbins):
  f.write('%.5f  '%xg[i] +'%.5f  '%yg[i] +'%.5f    '%xGa[i] +'%.5f  '%yGa[i] + '\n')
f.close()


# Plot the results
# Comment the following if matplotlib is not installed
matplotlib.rcParams['xtick.major.pad'] = 8
matplotlib.rcParams['ytick.major.pad'] = 8

fig = plt.figure(figsize=(20,7))
ax = fig.add_subplot(121)

plt.plot(xGa,yGa,label=r'Lilith')
plt.ylim([0,6])
plt.xlim([0.5,2])
plt.xlabel(r'$C_\gamma$', fontsize=22)
plt.ylabel(r'$\Delta \chi^2$', fontsize=22)
plt.axhline(y=1.,color='g',ls='dashed')
plt.axhline(y=4.,color='g',ls='dashed')
plt.axhline(y=9.,color='g',ls='dashed')

ax = fig.add_subplot(122)

plt.plot(xg,yg,label=r'Lilith')
plt.ylim([0,6])
plt.xlim([0.5,1.6])
plt.xlabel(r'$C_g$', fontsize=22)
plt.ylabel(r'$\Delta \chi^2$', fontsize=22)
plt.axhline(y=1.,color='g',ls='dashed')
plt.axhline(y=4.,color='g',ls='dashed')
plt.axhline(y=9.,color='g',ls='dashed')


plt.legend(prop={'size':16}, loc=2)
plt.savefig(outputfile+".pdf", bbox_inches='tight')
