#########################################################
# Lilith
#
# Routine example to perform a coupling fit
# Use the Minuit library iminuit to perform minimization
# Model: (C_V, C_F)
#
#########################################################

import lilith, tools.genXML2
from iminuit import Minuit

# Use library matplotlib to plot the results
# To comment if not installed
from pylab import *
import matplotlib.pyplot as plt


######################################################################

# Choose which experimental data to fit
myexpinput = "data/CMS14009.list"
# Lilith precision mode
myprecision = "BEST-QCD"
# Higgs mass to test
myhmass = 125.5
# Output file name (plots + dat file)
outputfile = 'results/1D_CMS_CVCF'
verbose=False

######################################################################



######################################################################

def getL(CF, CV):
  """ Returns -2Log(L) for given (C_V, C_F), based on Lilith classes and functions """
  myXML_user_input = tools.genXML2.redC(myhmass, Ctt=CF, Cbb=CF, Ctautau=CF, Ccc=CF, CWW=CV, CZZ=CV, precision=myprecision)
  l = Lilithcalc.GetLikelihood(myXML_user_input)[0]
  return l

######################################################################

# Initialize a Lilith object
Lilithcalc = lilith.Lilith(verbose)
# Read experimental data
Lilithcalc.SetExpInput(myexpinput)

# Initialize the fit; parameter starting values and limits
m = Minuit(getL, CV=0.9, limit_CV=(0,2), CF=0.9, limit_CF=(0,2))
print m.print_param()
# Fit the model
m.migrad()
# Display parameter values at the best-fit point
print "C_gauge:", m.values["CV"], " C_fermions:", m.values["CF"], " -2Log(L) min:", m.fval

# Get the 1D profile for the parameters
nbins = 250
xV,yV,rV = m.mnprofile('CV', bins=nbins, bound=(0.0, 2.0), subtract_min=True)
xF,yF,rF = m.mnprofile('CF', bins=nbins, bound=(0.0, 2.0), subtract_min=True)


# Save results in output file
f=open(outputfile+".dat",'w')
for i in range(1,len(xV)):
  f.write('%.5f  '%xV[i] +'%.5f  '%yV[i] +'%.5f    '%xF[i] +'%.5f  '%yF[i] + '\n')
f.close()


# Plot the results
# Comment the following if matplotlib is not installed
matplotlib.rcParams['xtick.major.pad'] = 8
matplotlib.rcParams['ytick.major.pad'] = 8

fig = plt.figure(figsize=(20,7))
ax = fig.add_subplot(121)

plt.plot(xV,yV,label=r'Lilith')
plt.ylim([0,6])
plt.xlim([0,2])
plt.xlabel(r'$C_V$', fontsize=22)
plt.ylabel(r'$\Delta \chi^2$', fontsize=22)
plt.axhline(y=1.,color='g',ls='dashed')
plt.axhline(y=4.,color='g',ls='dashed')
plt.axhline(y=9.,color='g',ls='dashed')

ax = fig.add_subplot(122)

plt.plot(xF,yF,label=r'Lilith')
plt.ylim([0,6])
plt.xlim([0,2])
plt.xlabel(r'$C_F$', fontsize=22)
plt.ylabel(r'$\Delta \chi^2$', fontsize=22)
plt.axhline(y=1.,color='g',ls='dashed')
plt.axhline(y=4.,color='g',ls='dashed')
plt.axhline(y=9.,color='g',ls='dashed')


plt.legend(prop={'size':16}, loc=2)
plt.savefig(outputfile+".pdf", bbox_inches='tight')
