import sys
import xml.etree.ElementTree as ET
from internal.ErrorHandling import *
from scipy import interpolate
import numpy as np
import math

class ReadExpInput(ErrorHandling):
    """Read the experimental input in XML and extracts all information."""

    def __init__(self):
        """..."""
    
        self.mu = []

    def get_filelist(self, list_input_filepath):
        """get list of files from .list experimental input"""
        expfiles = [] # list of all paths to XML input files

        filepath_split = list_input_filepath.split("/")
        expdata_dir = "/".join(filepath_split[:-1])

        try:
            linput = open(list_input_filepath)
        except IOError as e:
            print "I/O error({0}): {1}".format(e.errno, e.strerror)
            print "Cannot open the input list file \"" + list_input_filepath + "\"."
            sys.exit()
        expfiles = [] # list of all paths to XML input files

        for line in linput:
            # remove comment, space and new lines from the line
            line = line.split("#")[0].rstrip("\n").strip()
            if line == "": # empty line or comment
                continue
            if line[0] == "/": # absolute filepath
                expfiles.append(line)
            else: # relative filepath
                expfiles.append(expdata_dir+"/"+line)

        linput.close()
        
        return expfiles
    

    def read_file(self, filepath, effWZ):
        """read individual xml files"""

        root = self.produce_tree(filepath)
        if root.tag != "expmu":
            self.error("root tag is not <expmu>")

        self.root = root

        (dim, decay, type) = self.get_mode()
        
        self.get_mh()
        (experiment, source, sqrts) = self.get_metadata()
        
        eff = self.read_eff(dim, decay, effWZ)
        (bestfit, param, grid, Lxy, LChi2min) = self.read_mus(dim, type)

        self.mu.append({"filepath": filepath,
                         "dim": dim, "type": type,
                         "eff": eff,
                         "bestfit": bestfit, "param": param, "grid": grid,
                         "Lxy": Lxy, "LChi2min": LChi2min,
                         "experiment": experiment, "source": source,
                         "sqrts": sqrts, "eff": eff})

        del self.root

    def produce_tree(self, filepath):
        """Produce the XML tree with ElementTree."""

        try:
            f = open(filepath)
        except IOError as e:
            print "I/O error({0}): {1}".format(e.errno, e.strerror)
            print "Cannot open the experimental input file \"" + filepath + "\"."
            sys.exit()

        tree = ET.parse(f)
        f.close()
        return tree.getroot()

    def get_mode(self):
        """Get the dimension, decay and type of the experimental mu."""
        
        allowed_decays = ["gammagamma", "ZZ", "WW", "Zgamma",
                          "tautau", "bb", "cc", "invisible"]

        mandatory_attribs = {"dim":["1", "2"],
                             "type":["n", "f"]}
        optional_attribs = {"decay": allowed_decays}

        for mandatory_attrib, allowed_values in mandatory_attribs.items():
            if mandatory_attrib not in self.root.attrib:
                # if "dim" or "type" not in attribute
                self.error("mandatory attribute of root tag \"" +
                mandatory_attrib + "\" is not present.")
            
            if self.root.attrib[mandatory_attrib] not in allowed_values:
                # if dim="3" or type="z" for instance
                self.error("mandatory attribute of root tag \"" +
                mandatory_attrib + "\" has value \"" +
                self.root.attrib[mandatory_attrib] + "\" which is unknown. Allowed values are : " + str(allowed_values))

        dim = int(self.root.attrib["dim"])
        type = self.root.attrib["type"]

        decay = "mixture"

        for optional_attrib, allowed_values in optional_attribs.items():
            if optional_attrib in self.root.attrib:
                # if "decay" in attribute
                if self.root.attrib[optional_attrib] not in allowed_values:
                    # if decay="yy" for instance
                    self.error("optional attribute of root tag \"" +
                    optional_attrib + "\" has value \"" +
                    self.root.attrib[optional_attrib] + "\" which is unknown. Allowed values are : " + str(allowed_values))
                else:
                    decay = self.root.attrib["decay"]

        return (dim, decay, type)
    
    def get_mh(self):
        def_mh = 125.5 # default value
        mh = def_mh
    
        for child in self.root:
            if child.tag == "mh":
                try:
                    mh = float(child.text)
                except TypeError: # empty tag is of type NULL
                    mh = def_mh
                except ValueError:
                    self.error("value of <mh> tag is not a number")
        self.mh = mh

    def get_metadata(self):
        experiment = ""
        source = ""
        sqrts = ""

        for child in self.root:
            if child.tag == "experiment":
                experiment = child.text
            if child.tag == "source":
                source = child.text
            if child.tag == "sqrts":
                sqrts = child.text
    
        return (experiment, source, sqrts)

    def read_eff(self, dim, decay, eff_VVH):

        allowed_decays = ["gammagamma", "ZZ", "WW", "Zgamma",
                          "tautau", "bb", "cc", "invisible"]

        # read the efficiencies
        if dim == 1: # 1D signal strength
            eff = {"x": {}}
            axis_label = "x"

            mandatory_attribs = {"prod": ["ggH", "VVH", "VBF", "VH", "WH", "ZH", "ttH"]}
            if decay == "mixture":
                mandatory_attribs["decay"] = allowed_decays

            for child in self.root:
                if child.tag == "eff":
                    for mandatory_attrib, allowed_values in mandatory_attribs.items():
                        if mandatory_attrib not in child.attrib:
                            # if "axis" or "prod" not in attribute
                            self.error("mandatory attribute of <eff> tag \"" +
                            mandatory_attrib + "\" is not present.")
                        if child.attrib[mandatory_attrib] not in allowed_values:
                            # if axis="z" or prod="yy"
                            self.error("mandatory attribute of <eff> tag \"" +
                            mandatory_attrib + "\" has value \"" +
                            child.attrib[mandatory_attrib] + "\" which is unknown. Allowed values are : " + str(allowed_values))

                    prod_label = child.attrib["prod"]
                    if decay == "mixture":
                        decay_label = child.attrib["decay"]
                    else:
                        decay_label = decay

                    if prod_label in eff["x"]:
                        self.warning("<eff> tag with prod=\"" + prod_label + "\" is being redefined")
                        
                    try:
                        eff[axis_label][prod_label,decay_label] = float(child.text)
                    except TypeError: # empty tag is of type NULL
                        eff[axis_label][prod_label,decay_label] = 0.
                    except ValueError:
                        self.error('value of <eff> tag with axis="' + axis_label + \
                        '" and prod="' + prod_label + '" and decay="' + decay_label + '" is not a number')

        else: # 2D signal strength
            eff = {"x": {}, "y": {}}

            mandatory_attribs = {"axis": ["x", "y"],
                             "prod": ["ggH", "VVH", "VBF", "VH", "WH", "ZH", "ttH"]}
            if decay == "mixture":
                mandatory_attribs["decay"] = allowed_decays
            
            for child in self.root:
                if child.tag == "eff":
                    for mandatory_attrib, allowed_values in mandatory_attribs.items():
                        if mandatory_attrib not in child.attrib:
                            # if "axis" or "prod" not in attribute
                            self.error("mandatory attribute of <eff> tag \"" +
                            mandatory_attrib + "\" is not present.")
                        if child.attrib[mandatory_attrib] not in allowed_values:
                            # if axis="z" or prod="yy"
                            self.error("mandatory attribute of <eff> tag \"" +
                            mandatory_attrib + "\" has value \"" +
                            child.attrib[mandatory_attrib] + "\" which is unknown. Allowed values are : " + str(allowed_values))

                        axis_label = child.attrib["axis"]
                        prod_label = child.attrib["prod"]
                        if decay == "mixture":
                            decay_label = child.attrib["decay"]
                        else:
                            decay_label = decay


                    if prod_label in eff[axis_label]:
                        self.warning("<eff> tag with axis=\"" + axis_label + \
                        "\" and prod=\"" + prod_label + "\" is being redefined")
                        
                    try:
                        eff[axis_label][prod_label,decay_label] = float(child.text)
                    except TypeError: # empty tag is of type NULL
                        eff[axis_label][prod_label,decay_label] = 0.
                    except ValueError:
                        self.error('value of <eff> tag with axis="' + axis_label + \
                        '" and prod="' + prod_label + '" and decay="' + decay_label + '" is not a number')

        effWH = eff_VVH["eff_WH"](self.mh)
        effZH = eff_VVH["eff_ZH"](self.mh)
        
        effVBF = eff_VVH["eff_VBF"](self.mh)
        effVH = eff_VVH["eff_VH"](self.mh)
        effVWH = effVH * effWH
        effVZH = effVH * effZH

        multiprod = {"VH": {"WH": effWH, "ZH": effZH}, "VVH": {"VBF": effVBF, "WH": effVWH, "ZH": effVZH}}

        self.check_multiprod(eff["x"], multiprod)
        if dim == 2:
            self.check_multiprod(eff["y"], multiprod)

        # now all reduced couplings have been properly defined, one can
        # delete all multiparticle labels
        effCleanX = eff["x"].copy()
        for (p,decay) in eff["x"]:
            if p in multiprod:
                del effCleanX[p,decay]

        if dim == 2:
            effCleanY = eff["y"].copy()

            for (p,decay) in eff["y"]:
                if p in multiprod:
                    del effCleanY[p,decay]

        eff["x"] = effCleanX
        if dim == 2:
            eff["y"] = effCleanY

        # check that efficiencies add up to 1, otherwise issue a warning
        # or an error
        for axis in eff:
            sumeff = 0
            for prod in eff[axis]:
                sumeff += eff[axis][prod]
            
            if sumeff == 0:
                self.error("no <eff> tag found for " + axis + " axis")

            if sumeff < 0.99:
                self.warning("the sum of efficiencies for axis=\"" +
                             axis + "\" is less than 1 (value: " +
                             str(sumeff) + ")")
            elif sumeff > 1.01:
                self.error("the sum of efficiencies for axis=\"" +
                             axis + "\" is greater than 1 (value: " +
                             str(sumeff) + ")")

        return eff


    def read_mus(self, dim, type):
        # first, read the bestfit
        bestfit = {}
        LChi2min = 0
                
        for child in self.root:
            if child.tag == "bestfit":
                if type == "f":
                    self.warning("block <bestfit> in experimental mu of" +
                                 "type \"full\"... skipping.")
                    continue

                if dim == 1:
                    # read directly the number
                    
                    if "x" in bestfit:
                        self.warning("redefinition of the bestfit...")
                    
                    try:
                        bestfit["x"] = float(child.text)
                    except TypeError: # empty tag is of type NULL
                        bestfit["x"] = 0.
                    except ValueError:
                        self.error("value of <besfit> tag is not a number")
            
                elif dim == 2:
                    bestfit_allowedsubtags = ["x", "y"]
                
                    for bfit in child:
                        if bfit.tag in bestfit_allowedsubtags:
                            
                            if bfit.tag in bestfit:
                                self.warning("redefinition of the bestfit...")
                        
                            try:
                                bestfit[bfit.tag] = float(bfit.text)
                            except TypeError: # empty tag is of type NULL
                                bestfit[bfit.tag] = 0.
                            except ValueError:
                                self.error("value of <besfit> tag is not a number")
                        else:
                            self.error("subtag in bestfit not known")
                    
                if dim == 1 and "x" not in bestfit:
                    self.error("best fit point should be specified.")
                if dim == 2 and ("x" not in bestfit or "y" not in bestfit):
                    self.error("best fit point should be specified for x and y.")
        
        # then, read the param...
        param = {}
        
        for child in self.root:
            if child == "param":
                    break
        param_tag = child

        param["uncertainty"] = {}
        
        for child in param_tag:
            if dim == 1:
                if child.tag == "uncertainty":
                    if "side" not in child.attrib:
                        try:
                            unc_value = float(child.text)
                        except TypeError: # empty tag is of type NULL
                            unc_value = 0.
                        except ValueError:
                            self.error("value of <uncertainty> tag is not a number")

                        param["uncertainty"]["left"] = unc_value
                        param["uncertainty"]["right"] = unc_value
                    else:
                        if child.attrib["side"] not in ["left", "right"]:
                            self.error("attribute of uncertainty is not left nor right")
                        else:
                            try:
                                unc_value = float(child.text)
                            except TypeError: # empty tag is of type NULL
                                unc_value = 0.
                            except ValueError:
                                self.error("value of <uncertainty> tag is not a number")
                                
                        param["uncertainty"][child.attrib["side"]] = unc_value
                else:
                    self.error("subtag or param should be uncertainty")
                    
            elif dim == 2:
                allowed_tags = ["a", "b", "c"]
                if child.tag not in allowed_tags:
                    self.error("only allowed tags are <a>, <b> and <c> in block param in 2D normal mode")

                if child.tag in param:
                    self.warning("redefinition of tag <" + child.tag + ">")
                    
                try:
                    param_value = float(child.text)
                except TypeError: # empty tag is of type NULL
                    param_value = 0.
                except ValueError:
                    self.error("value of <" + child.tag + "> tag is not a number")
                
                param[child.tag] = param_value
        
        # check that everything is there
        if type == "n" and dim == 1:
            if "uncertainty" not in param or \
            "left" not in param["uncertainty"] or \
            "right" not in param["uncertainty"]:
                self.error("uncertainties are not given consistently in block param")
        elif type == "n" and dim == 2:
            if "a" not in param or "b" not in param or "c" not in param:
                self.error("a, b, c tags are not given in block param")
        
        # or the grid
        grid = {}
        Lxy = None
        
        for child in self.root:
            if child == "grid":
                break
        grid_raw = child.text

        if type == "f" and dim == 1:
            x = []
            L = []
        
            grid_raw = grid_raw.strip("\n").strip().split("\n")
        
            i = -1
        
            for line in grid_raw:
                tab = line.split()
                if len(tab) != 2:
                    self.error("incorrect <grid> entry on line \"" + line + "\"")

                cur_x = float(tab[0])
                cur_L = float(tab[1])

                if cur_x not in x:
                    x.append(cur_x)
                    L.append(cur_L)
                    i += 1
                else:
                    i = x.index(cur_x)

            grid["x"] = x
            grid["L"] = L
            LChi2min = min(grid["L"])

            Lxy = interpolate.UnivariateSpline(grid["x"], grid["L"], k = 3, s = 0)
        
        elif type == "f" and dim == 2:
            x = []
            y = []
            L = []
            
            grid_raw = grid_raw.strip("\n").strip().split("\n")
            
            i = -1
            
            for line in grid_raw:
                tab = line.split()
                if len(tab) != 3:
                    self.error("incorrect <grid> entry on line \"" + line + "\"")

                cur_x = float(tab[0])
                cur_y = float(tab[1])
                cur_L = float(tab[2])

                if cur_x not in x:
                    x.append(cur_x)
                    L.append([])
                    i += 1
                else:
                    i = x.index(cur_x)

                if cur_y not in y:
                    y.append(cur_y)

                L[i].append(cur_L)

            grid["x"] = np.array(x)
            grid["y"] = np.array(y)
            grid["L"] = np.array(L)
            
            LChi2min = min(min(p[1:]) for p in grid["L"])

            Lxy = interpolate.RectBivariateSpline(grid["x"],
                    grid["y"], grid["L"])
            
        return (bestfit, param, grid, Lxy, LChi2min)

    def check_multiprod(self, eff_dict, multiprod):
        """..."""

        # check consistency in the definition of multi-particle labels
        for (prod,decay) in eff_dict:
            if prod in multiprod:
                # there is a multi-particle label
                # in that case, check if individual particle are also defined
                for label in multiprod[prod]:
                    if (label,decay) in eff_dict:
                        self.error('<eff> tags for "' + label + '" and "' + \
                        prod + '" cannot both be defined')
                # also, only one multi-particle label can be used (VH or VVH),
                # not both
                for label in multiprod:
                    if label != prod and (label,decay) in eff_dict:
                        self.error('<eff> tags for "' + label + '" and "' + \
                        prod + '" cannot both be defined')

        # it is consistent, resolve multi-particle labels
        new_eff = {}
        for (prod,decay) in eff_dict:
            if prod in multiprod:
                for label in multiprod[prod]:
                    new_eff[label,decay] = eff_dict[prod,decay]*multiprod[prod][label]

        for elem in new_eff:
            eff_dict[elem] = new_eff[elem]
