import sys
import os
try:
    from lxml import etree
except:
    import xml.etree.ElementTree as etree
from internal.ErrorHandling import *



class ReadUserInput(ErrorHandling):
    """Read the XML input and extracts all information."""

    def __init__(self, inputstring):
        """Initialize the reading of the user input from the XML input contained
        in the string inputstring."""

        self.inputstring = inputstring

        try:
            tree = etree.parse(self.inputstring)
            self.root = tree.getroot()
        except:
            self.root = etree.fromstring(self.inputstring)
            #print "Input file not found in ", self.inputstring
    
        if self.root.tag != "lilithinput":
            self.error('root tag is not <lilithinput>')

        self.mode = self.get_mode()
        self.mh, self.isInLightRange = self.get_mh()

        self.redC = {}
        self.extraBR = {} # only filled when mode is "reducedcouplings"
        self.mu = {}
        self.precision = "" # only filled when the mode is "reducedcouplings"

        if self.mode == "reducedcouplings":
            self.read_reducedcouplings()
        elif self.mode == "signalstrengths":
            self.read_signalstrengths()


    def produce_tree(self):
        """Produce the XML tree with ElementTree."""

        return etree.fromstring(self.inputstring)
    

    def get_mode(self):
        """Get the mode (signal strengths or reduced couplings) from the
           XML tree."""

        modes = ["signalstrengths", "reducedcouplings"]
        curmode = ""

        for child in self.root:
            if child.tag in modes:
                # check if no more than one mode is present in the XML file
                if curmode != "":
                    self.error('<' + child.tag + '> and <' + curmode + \
                          '> blocks cannot both be present')
                curmode = child.tag
    
        # check if at least one mode is present
        if curmode == "":
            self.error('no mode block has been defined (' + \
                  '<signalstrengths> or <reducedcouplings>)')
        
        return curmode
    
    def get_mh(self):
        """Get the Higgs mass (in GeV) from the XML tree."""
        
        mh = -1.
        default_mh = 125.5
        min_mh = 123.
        max_mh = 128.

        for child in self.root:
            if child.tag == "mh":
                if mh > 0:
                    self.warning('redefinition of mh ...')
                try:
                    mh = float(child.text)
                    light = True
                    assert mh >= min_mh and mh <= max_mh
                except TypeError: # empty tag is of type NULL
                    mh = default_mh
                    self.warning('<mh> tag is empty; setting to ' +
                                 'SM value=125.5 GeV')
                except ValueError:
                    self.error('value of the <mh> tag is not a number.')
                except AssertionError:
                   self.warning('<mh> is not between 123 and 128 GeV...') #!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                   light = False
                break

        # if no <mh> block, set to default value
        if mh < 0:
            mh = default_mh

        return mh, light

    def read_reducedcouplings(self):
        """..."""
        
        accepted_C = ["ff", "uu", "tt", "cc", "dd", "bb", "tautau", "VV", "WW", \
                      "ZZ", "gg", "gammagamma", "Zgamma","VBF"]
        accepted_precision = ["LO", "BEST-QCD"]
        accepted_extraBR = ["invisible", "undetected"]

        for child in self.root:
            if child.tag == "reducedcouplings":
                red_coupl = child
            elif child.tag == "extraBR":
                extra_BR = child
                break
        
        for child in red_coupl:
            if child.tag == "C":
                if "to" not in child.attrib:
                    self.warning('attribute "type" is missing in <C> tag')
                elif child.attrib["to"] not in accepted_C:
                    self.warning('<C> tag to "' + child.attrib["to"] + \
                            '" is unknown')
                else:
                    if child.attrib["to"] == "gg":
                        # gg has optional attribute for="prod|decay|all"
                        # to be improved...
                        if "for" in child.attrib:
                            if child.attrib["for"] in ["prod", "decay", "all"]:
                                C_to = "gg_" + child.attrib["for"]
                                
                                if child.attrib["for"] == "prod":
                                    # gg for="prod" has optional attribute
                                    # at=lhc8|tevatron|all
                                    if "at" in child.attrib:
                                        if child.attrib["at"] in ["lhc8", "tevatron", "all"]:
                                            C_to = "gg_prod_" + child.attrib["at"]
                                        else:
                                            self.error('<C to="gg" for="prod">' +
                                            ' and at="' + child.attrib["at"] +
                                            '" is unknown...')
                                    else:
                                        C_to = "gg_prod_all"
                            else:
                                self.error('<C to="gg"> for "' +
                                child.attrib["for"] + '" is unknown...')
                        else:
                            C_to = "gg_all"
                    else:
                        C_to = child.attrib["to"]
                    
                    if child.attrib["to"] in ["ff", "uu", "tt", "cc", "dd", "bb", "tautau"]:
                        # if fermion, there could be a real and imaginary part
                        if "part" in child.attrib:
                            if child.attrib["part"] in ["re", "im"]:
                                C_to += "_" + child.attrib["part"]
                    
                    if C_to in self.redC:
                        self.warning('reduced coupling for "' + \
                        C_to + '" is being redefined')
                    try:
                        self.redC[C_to] = float(child.text)
                    except TypeError: # empty tag is of type NULL
                        self.redC[C_to] = 1.
                        self.warning('<C> tag to "' + child.attrib["to"] + \
                                '" is empty; setting to SM value = 1')
                    except ValueError:
                        self.error('value of the <C> tag to "' +
                              child.attrib["to"] + '" is not a number.')
        
            if child.tag == "precision":
                if child.text not in accepted_precision:
                    self.warning('"' + str(child.text) +
                            '" precision is not allowed; setting to default ' +
                            'value = "BEST-QCD"')
                    self.precision = "BEST-QCD"
                else:
                    if self.precision != "":
                        self.warning('precision is being redefined')
                    self.precision = child.text
                break

        # --- putting together real and imaginary part when present
        new_redC = {}
        original_redC = {}
        for p in self.redC:
            new_p = 0.
            if p[-3:] in ["_im", "_re"]:
                if p[:-3] in self.redC:
                    self.error(p[:-3] + ' and ' + p + ' are defined at the same time')
                
                if p[-3:] == "_re":
                    num = self.redC[p]
                else:
                    num = self.redC[p]*1j
                
                if p[:-3] in new_redC:
                    new_redC[p[:-3]] += num
                else:
                    new_redC[p[:-3]] = num
            else:
                original_redC[p] = self.redC[p]

        self.redC = original_redC
        self.redC.update(new_redC)

        # ----------------------------------
        # checking consistency of the input
        # ----------------------------------

        # --- checking the multiparticles
        multiparticles = {"VV": ["WW", "ZZ"], "uu": ["tt", "cc"], \
                          "dd": ["bb", "tautau"]}
        multiparticles2 = {"ff": ["tt", "cc", "bb", "tautau"]}
        # two separate dicts because "ff" is a special case since it overlaps
        # with the multiparticle labels "uu" and "dd"---it should be treated
        # in the very end

        for multip, p_list in multiparticles.items():
            self.check_multiparticle(multip, p_list)
        
        for multip, p_list in multiparticles2.items():
            self.check_multiparticle(multip, p_list)

        # now all reduced couplings have been properly defined, one can
        # delete all multiparticle labels
        redCclean = self.redC.copy()
        for p in self.redC:
            if p in multiparticles or p in multiparticles2:
                del redCclean[p]
        self.redC = redCclean

        # --- checking that the couplings to all particles are well defined in
        #     input (not including optionnal couplings for the the loop-induced
        #     processes)
        particles = ["WW", "ZZ", "tt", "cc", "bb", "tautau"]

        for p in particles:
            if p not in self.redC:
                #self.warning('reduced coupling to "' + p + '" is not specified; fixing it to SM value=1')
                self.redC[p] = 1.

        # --- checking that the couplings to "gg" (for="prod|decay|all") are
        #     consistently defined
        # also, when for="prod" there is an argument at="lhc8|tevatron|all"

        if "gg_prod_all" in self.redC:
            if "gg_prod_lhc8" in self.redC and "gg_prod_tevatron" in self.redC:
                self.warning('gg_prod_all, gg_prod_lhc8 and gg_prod_tevatron defined at the same time. ignoring gg_prod_all')
            elif "gg_prod_lhc8" in self.redC:
                self.warning('only gg_prod_lhc8 and gg_prod_all defined... taking gg_prod_tevatron = gg_all')
                self.redC["gg_prod_tevatron"] = self.redC["gg_prod_lhc8"]
            elif "gg_prod_tevatron" in self.redC:
                self.warning('only gg_prod_tevatron and gg_prod_all defined... taking gg_prod_lhc8 = gg_all')
                self.redC["gg_prod_lhc8"] = self.redC["gg_prod_tevatron"]
            else:
                self.redC["gg_prod_lhc8"] = self.redC["gg_prod_all"]
                self.redC["gg_prod_tevatron"] = self.redC["gg_prod_all"]

            del self.redC["gg_prod_all"]
        else:
            if "gg_prod_lhc8" in self.redC and "gg_prod_tevatron" not in self.redC:
                self.warning('only gg_prod_lhc8 defined... taking gg_prod_tevatron = gg_gg_prod_lhc8')
                self.redC["gg_prod_tevatron"] = self.redC["gg_prod_lhc8"]
            elif "gg_prod_tevatron" in self.redC and "gg_prod_lhc8" not in self.redC:
                self.warning('only gg_prod_tevatron defined... taking gg_prod_lhc8 = gg_prod_tevatron')
                self.redC["gg_prod_lhc8"] = self.redC["gg_prod_tevatron"]

        # if for="prod" and at="something" have been given, there is
        # gg_prod_lhc8 and gg_prod_tevatron

        if "gg_all" in self.redC:
            if "gg_prod_lhc8" in self.redC and "gg_decay" in self.redC:
                self.warning('gg_all, gg_prod and gg_decay defined at the same time. ignoring gg_all')
            elif "gg_prod_lhc8" in self.redC:
                self.warning('only gg_prod and gg_all defined... taking gg_decay = gg_all')
                self.redC["gg_decay"] = self.redC["gg_all"]
            elif "gg_decay" in self.redC:
                self.warning('only gg_decay and gg_all defined... taking all gg_prod = gg_all')
                self.redC["gg_prod_lhc8"] = self.redC["gg_all"]
                self.redC["gg_prod_tevatron"] = self.redC["gg_all"]
            else:
                self.redC["gg_prod_lhc8"] = self.redC["gg_all"]
                self.redC["gg_prod_tevatron"] = self.redC["gg_all"]
                self.redC["gg_decay"] = self.redC["gg_all"]

            del self.redC["gg_all"]
        else:
            if "gg_prod_lhc8" in self.redC and "gg_decay" not in self.redC:
                self.warning('gg prod has been defined, not gg decay... setting it to gg prod.')
                self.redC["gg_decay"] = self.redC["gg_prod_lhc8"]
            elif "gg_decay" in self.redC and "gg_prod_lhc8" not in self.redC:
                self.warning('gg decay has been defined, not any gg prod... setting them to gg decay.')
                self.redC["gg_prod_lhc8"] = self.redC["gg_decay"]
                self.redC["gg_prod_tevatron"] = self.redC["gg_decay"]
            # else: "gg" not defined at all, and it will be calculated a
            # posteriori from the other couplings

        # --- checking precision
        if self.precision == "":
            #self.warning('precision has not been specified; setting to default value = "BEST-QCD"')
            self.precision = "BEST-QCD"

        # --- checking <extraBR> block
        for child in extra_BR:
            if child.tag == "BR":
                if "type" not in child.attrib:
                    self.warning('attribute "type" is missing in <BR> tag')
                elif child.attrib["type"] not in accepted_extraBR:
                    self.warning('<BR> tag for type "' + child.attrib["type"] + \
                                 '" is unknown.')
                else:
                    if child.attrib["type"] in self.extraBR:
                        self.warning('<BR> tag with type ' + child.attrib["type"] +
                                     ' is being redefined')
                    
                    try:
                        self.extraBR[child.attrib["type"]] = float(child.text)
                    except TypeError: # empty tag is of type NULL
                        self.extraBR[child.attrib["type"]] = 0.
                    except ValueError:
                        self.error('value of the <BR> tag for "' +
                           child.attrib["type"] + '" is not a number.')
        
    def read_signalstrengths(self):
        """..."""

        accepted_prod = ["ggH", "VVH", "ttH", "VBF", "VH", "WH", "ZH"]
        accepted_decay = ["gammagamma", "VV", "WW", "ZZ", "bb", "tautau", \
                          "dd", "uu", "cc", "ff", "Zgamma", "mumu", "invisible"]

        for child in self.root:
            if child.tag == "signalstrengths":
                mu_tag = child
                break
    
        for child in mu_tag:
            if child.tag == "mu":
                if "prod" not in child.attrib or "decay" not in child.attrib:
                    self.warning('attribute "prod" or "decay" is ' + \
                                 'missing in <mu> tag')
                elif child.attrib["prod"] not in accepted_prod:
                    self.warning('<mu> tag with prod="' +
                                 child.attrib["prod"] + '" and decay="' + \
                                 child.attrib["decay"] + '" has unknown prod')
                elif child.attrib["decay"] not in accepted_decay:
                    self.warning('<mu> tag with prod="' +
                                 child.attrib["prod"] + '" and decay="' + \
                                 child.attrib["decay"] + '" has unknown decay')
                else:
                    prod = child.attrib["prod"]
                    decay = child.attrib["decay"]

                    if (prod,decay) in self.mu:
                        self.warning('<mu> tag with prod="' + prod + \
                                     '" and decay="' + decay + '" is being ' + \
                                     'redefined')

                    try:
                        self.mu[prod,decay] = float(child.text)
                    except TypeError: # empty tag is of type NULL
                        self.mu[prod, decay] = 1.
                        self.warning('<mu> tag with prod="' +
                                 prod + '" and decay="' + \
                                 decay + '" is empty; ' + \
                                 'setting to SM value=1')
                    except ValueError:
                        self.error('value of the <mu> tag with prod="' +
                                 prod + '" and decay="' + \
                                 decay + '" is not a number.')
            elif child.tag == "redxsBR":
                if "prod" not in child.attrib or "decay" not in child.attrib:
                    self.warning('attribute "prod" or "decay" is ' + \
                                 'missing in <redxsBR> tag')
                elif child.attrib["prod"] not in accepted_prod:
                    self.warning('<mu> tag with prod="' +
                                 child.attrib["prod"] + '" and decay="' + \
                                 child.attrib["decay"] + '" has unknown prod')
                elif child.attrib["decay"] != "invisible":
                    self.warning('<redxsBR> tag with prod="' +
                                 child.attrib["prod"] + '" and decay="' + \
                                 child.attrib["decay"] + '" has unknown decay')
                else:
                    prod = child.attrib["prod"]
                    decay = child.attrib["decay"]

                    if (prod,decay) in self.mu:
                        self.warning('<redxsBR> tag with prod="' + prod + \
                                     '" and decay="' + decay + '" is being ' + \
                                     'redefined')

                    try:
                        self.mu[prod,decay] = float(child.text)
                    except TypeError: # empty tag is of type NULL
                        self.mu[prod, decay] = 1.
                        self.warning('<redxsBR> tag with prod="' +
                                 prod + '" and decay="' + \
                                 decay + '" is empty; ' + \
                                 'setting to SM value=1')
                    except ValueError:
                        self.error('value of the <mu> tag with prod="' +
                                 prod + '" and decay="' + \
                                 decay + '" is not a number.')
        
        # ----------------------------------
        # checking consistency of the input
        # ----------------------------------

        # --- checking the multiparticle labels
        multiprod = {"VH": ["WH", "ZH"]}
        multiprod2 = {"VVH": ["VBF", "WH", "ZH"]}
        multidecay = {"VV": ["WW", "ZZ"], "uu": ["cc"], "dd": ["bb", "tautau"]}
        multidecay2 = {"ff": ["tt", "cc", "bb", "tautau"]}
        # two separate dicts for prod and decay because "ff" is a special case
        # since it overlaps with the multiparticle labels "uu" and "dd"
        # also "VVH" includes "VH"; this should be treated at the end

        # -- first: production
        for multip, p_list in multiprod.items():
            self.check_multiprod(multip, p_list)

        for multip, p_list in multiprod2.items():
            self.check_multiprod(multip, p_list)

        # now, one can delete all multiprod labels
        for (prod,decay), mu_value in self.mu.items():
            if prod in multiprod or prod in multiprod2:
                del self.mu[prod,decay]

        # -- then: decay
        for multip, p_list in multidecay.items():
            self.check_multidecay(multip, p_list)

        for multip, p_list in multidecay2.items():
            self.check_multidecay(multip, p_list)

        # now, one can delete all multidecay labels
        for (prod,decay), mu_value in self.mu.items():
            if decay in multidecay or decay in multidecay2:
                del self.mu[prod,decay]

        # --- checking that all mandatory signal strengths are present in the
        #     input
        mandatory_prod = ["ggH", "VBF", "WH", "ZH"]
        mandatory_decay = ["gammagamma", "ZZ", "WW", "bb", "tautau"]

        mandatory_mus = []
        for prod in mandatory_prod:
            for decay in mandatory_decay:
                mandatory_mus.append((prod,decay))

        for (prod,decay) in mandatory_mus:
            if (prod,decay) not in self.mu:
                self.warning('signal strength for prod "' + prod + \
                '" and decay "' + decay + '" is not  specified; fixing ' + \
                'it to SM value=1')
                self.mu[prod,decay] = 1.
                break
        
        # --- checking that invisible redxsBR are initialized to 0 if absent
        for prod in ["ggH", "VBF", "WH", "ZH", "ttH"]:
            if (prod,"invisible") not in self.mu:
                self.mu[prod,"invisible"] = 0.
        

    def check_multiparticle(self, multip, p_list):
        """for reduced couplings"""
        
        if multip in self.redC:
            if set(p_list).issubset(self.redC):
                # if multiparticle and individual particles are all
                # defined, check consistency
                for p in p_list:
                    if self.redC[multip] != self.redC[p]:
                        self.warning('inconsistent definition of the ' + \
                                'couplings to "' + multip + '" and "' + \
                                p + '"; skipping " + "<C to="' + \
                                multip + '"> tag')
                        break
            else:
                # only multiparticle label is present, or multiparticle
                # label and tags for only part of the associated particles
                given_p = [part for part in p_list if part in self.redC]
                
                # check if a particle tag is given inconsistently
                for p in given_p:
                    if self.redC[multip] != self.redC[p]:
                        self.error('inconsistent definition of the ' + \
                              'couplings to "' + multip + '" and to "' +
                              p + '"')
                        break
                
                if len(given_p) == 0:
                    # only multiparticle tag is defined
                    for p in p_list:
                        self.redC[p] = self.redC[multip]
                else:
                    # when multiparticle tag and part of the particle tags
                    # are given, with equal values
                    other_p = [part for part in p_list if \
                               part not in self.redC]
                    for op in other_p:
                        self.redC[op] = self.redC[multip]
                    self.warning('couplings to ' + multip + ' and ' + \
                            str(given_p) + ' are both defined and ' + \
                            'equal; assuming that the couplings to ' + \
                            str(other_p) + ' are the same')

    def check_multiprod(self, multip, p_list):
        """for signal strengths mode"""

        for (prod,decay), mu_value in self.mu.items():
            if prod == multip:
                if set([(subprod,decay) for subprod in p_list]).issubset(self.mu):
                    # if multiprod and individual prod are all
                    # defined, check consistency
                    for subprod in p_list:
                        if mu_value != self.mu[subprod,decay]:
                            self.warning('inconsistent definition of the ' + \
                            'mu prod="' + multip + '" and "' + \
                            prod + '" when decay="' + decay + '"; ' + \
                            'skipping ' + 'mu prod="' + multip + '" tag')
                            break
                else:
                    # only multiprod label is present, or multiprod
                    # label and tags for only part of the associated prod
                    given_mu = [(subprod, decay) for subprod in p_list if \
                                (subprod, decay) in self.mu]
                    
                    # check if a particle tag is given inconsistently
                    for (subprod,decay) in given_mu:
                        if mu_value != self.mu[subprod,decay]:
                            self.error('inconsistent definition of the ' + \
                                        'mu for prod "' + subprod + '" and "' + \
                                        multip + '" when decay is "' + decay + \
                                        '"')
                        break
            
                    if len(given_mu) == 0:
                        # only multiparticle tag is defined
                        for subprod in p_list:
                            self.mu[subprod,decay] = self.mu[multip,decay]
                    else:
                        # when multiprod tag and part of the prod tags
                        # are given, with equal values
                        other_mu = [(subprod, decay) for subprod in p_list if \
                                    (subprod, decay) not in self.mu]
                        for (op,decay) in other_mu:
                            self.mu[op,decay] = self.mu[multip,decay]
                        self.warning('signal strengths with prod ' + multip + \
                                     ' and ' + \
                                     str(given_mu) + ' are all defined and ' + \
                                     'equal; assuming that the signal strengths to '+ \
                                     str(other_mu) + ' are the same [for decay = "' + \
                                     decay + '"]')

    def check_multidecay(self, multip, p_list):
        """for signal strengths mode"""

        for (prod,decay), mu_value in self.mu.items():
            if decay == multip:
                if set([(prod,subdecay) for subdecay in p_list]).issubset(self.mu):
                    # if multidecay and individual decay are all
                    # defined, check consistency
                    for subdecay in p_list:
                        if mu_value != self.mu[prod,subdecay]:
                            self.warning('inconsistent definition of the ' + \
                            'mu decay="' + multip + '" and "' + \
                            decay + '" when prod="' + prod + '"; ' + \
                            'skipping ' + 'mu decay="' + multip + '" tag')
                else:
                    # only multidecay label is present, or multidecay
                    # label and tags for only part of the associated prod
                    given_mu = [(prod, subdecay) for subdecay in p_list if \
                                (prod, subdecay) in self.mu]
                    
                    # check if a particle tag is given inconsistently
                    for (prod,subdecay) in given_mu:
                        if mu_value != self.mu[prod,subdecay]:
                            self.error('inconsistent definition of the ' + \
                                        'mu for decay "' + subdecay + '" and "' + \
                                        multip + '" when prod is "' + prod + \
                                        '"')
            
                    if len(given_mu) == 0:
                        # only multiparticle tag is defined
                        for subdecay in p_list:
                            self.mu[prod,subdecay] = self.mu[prod,multip]
                    else:
                        # when multidecay tag and part of the decay tags
                        # are given, with equal values
                        other_mu = [(prod, subdecay) for subdecay in p_list if \
                                    (prod, subdecay) not in self.mu]
                        for (prod,od) in other_mu:
                            self.mu[prod,od] = self.mu[prod,multip]
                        self.warning('signal strengths with decay ' + multip + \
                                     ' and ' + \
                                     str(given_mu) + ' are all defined and ' + \
                                     'equal; assuming that the signal strengths to '+ \
                                     str(other_mu) + ' are the same [for prod = "' + \
                                     prod + '"]')
